@@ -6,7 +6,6 @@
 #include <boost/test/unit_test.hpp>
 
 #include "atomic_queue/atomic_queue.h"
-#include "atomic_queue/atomic_queue_mutex.h"
 #include "atomic_queue/barrier.h"
 
 #include <cstdint>
@@ -250,6 +249,8 @@ BOOST_AUTO_TEST_CASE(try_push) {
 BOOST_AUTO_TEST_CASE(size) {
     atomic_queue::RetryDecorator<atomic_queue::AtomicQueueB2<float>> q(10);
     BOOST_CHECK_EQUAL(q.capacity(), CACHE_LINE_SIZE * CACHE_LINE_SIZE);
+    BOOST_CHECK(q.was_empty());
+    BOOST_CHECK(!q.was_full());
 }
 
 BOOST_AUTO_TEST_CASE(power_of_2) {
